unit SimpleMail;

{*****************************************************************************
 *  Unit:  SimpleMail                                                        *
 *  Zweck: Versenden und Empfangen von e-Mails mit Daten, die an die Objekte *
 *         TSMTPObject oder TPOP3Object bergeben wurden                     *
 *  Autor: Jrgen Hummel   ***   post@hummel1.de   ***   www.hummel1.de      *
 *  Stand: 2001-05-25                                                        *
 *****************************************************************************

-------------------------------------------------------------------------------
TPOP3Object:
============
Eigenschaften:

  Server:     String                 Adresse des POP3-Servers
  Konto:      String                 Username/Account
  Pass:       String                 Passwort
  Checked:    Boolean                Pop3-Server wurde mit Erfolg kontaktiert
  Count:      Integer                Anzahl der Nachrichten auf dem POP3-Server

Methoden:

  CheckAccount:             Integer  POP-Server kontaktieren u. MailInfos holen
  GetSubject(n: Integer):   String   liefert Betreff der n-ten Mail
  GetFrom(n: Integer):      String   liefert Absender der n-ten Mail
  GetTo(n: Integer):        String   liefert Empfnger der n-ten Mail
  GetDate(n: Integer):      String   liefert Datum der n-ten Mail
  GetSize(n: Integer):      String   liefert Gre der n-ten Mail
  RetrMail(Nr: Integer):    Integer  n-te Mail vom POP-Server abrufen
  DeleteMails(Nrs: String): Integer  lscht die Mails laut Liste Nrs vom Server
                                     (Nrs ist eine CR/LF getrennte Augzhlung)
  GetHeader:                String   liefert Header der zuletzt abgeruf. Mail
  GetMailText:              String   liefert Text der zuletzt abgerufenen Mail
  GetAttachNames:           String   holt AnhangNamen der zuletzt abgeruf. Mail
  StoreAttachment(Nr: Integer; sFNa: TFileName): Boolean
                                     speichert Anhang Nr der abgerufenen Mail
  StoreAsEml(FNa: TFileName):    Boolean  speichert die zuletzt abgerufene Mail
                                          im *.eml-Format
  LoadFromEml(sFNa: TFileName):  Boolean  ldt eine gespeicherte Mail statt sie
                                          vom Server zu holen
  CheckFromEml(sFNa: TFileName): Boolean  MailInfos aus einer gespeicherten
                                          Mail holen

-------------------------------------------------------------------------------
TSMTPObject:
============

Eigenschaften:

  Server:        String                Adresse des SMTP-Servers
  MailTo:        String                e-Mail-Adresse des Empfngers
  From:          String                e-Mail-Adresse des Absenders
  Sender:        String                Real-Name des Absenders
  Subject:       String                Betreff der e-Mail
  ContentLines:  Integer               Anzahl der Textzeilen im e-Mail-Inhalt
  SaveBySend:    Boolean               legt fest, ob beim Senden eine lokal zu
                                       speichernde Kopie erzeugt wird
  SaveFile:      TFileName             Name der eventuellen Sendekopie

Methoden:

  SetContent(const TheText: String)    setzt den zu sendenden Nachrichtentext
  SendMail: Integer;                   kompletter Sendevorgang der Nachricht
  GetContentLine(n: Integer): String   ruft eine Zeile aus d. Nachrichtentext ab
  AddAttachment(const TheName: String) fgt einen weiteren Dateinamen zur Liste
                                       der Mail-Anhnge hinzu
  ClearAttachments                     leert die Liste der Anhangsnamen

-------------------------------------------------------------------------------
Hilfsfunktionen:
================

  function  HxCharToInt(c: Char): Byte   liefert Wert eines HEX-Ziffernzeichens

  function  HtmlMaskLtGt(const s: String): String
            maskiert die Zeichen "<" und ">" und " als HTML-Zeichen "&lt;" und "&gt;" und "&quot;"

  function  DeBase64(const sIn: String): String;
            BASE64-String decodieren

  procedure DeQuote(var sIn: String)
            wandelt einen "quoted-printable"-Text (RFC 2045, 1341) in ANSI-Normaltext

  function  DeQuode(const sIn: String): String;
            wandelt einen "=?-?B/Q?-codierten"-Text (RFC 2047, 1342) in ANSI-Normaltext

  function  URLCode(const sIn: String): String;
            kodiert einen String in's "URL-encoded"-Format (RFC 1866)

 *****************************************************************************}

interface

uses Classes, Sysutils, Winsock;

                        {Ergebniswerte der e-Mail-Funktionen }
const SOCK_CONNECT =0;  {Server konnektiert}
      SMTP_MAILOK  =0;  {Mail korrekt gesendet}
      POP3_MAILOK  =0;  {Postfach erfolgreich abgefragt}
      SMTP_TRANSERR=1;  {Server meldet Fehler beim bertragen der Mail}
      POP3_NOPASS  =1;  {Passwortfehler beim Einloggen in das Postfach}
      POP3_NOKONTO =2;  {Username fehlerhaft beim Login in das Postfach}
      SOCK_NOSERVER=3;  {Mail-Server nicht gefunden/nicht verfgbar}
      SOCK_NONET   =4;  {keine TCP/IP-Netzverbindung}
      SMTP_NOPARAMS=5;  {nicht gengend Parameter zum Senden}
      POP3_NOPARAMS=5;  {nicht gengend Parameter zum Abruf einer Mail}

      IPPORT_POP3=110;

type TPOP3Object = class(TObject)         {das POP3-Client-Objekt}
 private
  FPopSock: TSocket;
  FServer:  String;                       {Name des zu benutzenden POP3-Servers}
  FKonto:   String;                       {Konto des e-Mail-Empfngers}
  FPass:    String;                       {e-Mail-Passwort}
  FCount:   Integer;                      {Anzahl der Nachrichten in der Box}
  FFroms, FSubjects, FDates, FSizes: TStringList;        {Daten der Nachrichten}
  FChecked: Boolean;                      {Server wurde erfolgreich kontaktiert}
  function FindLine(nAt: Integer; s: String): Integer; {Zeile mit Anfang s find}
  function GetEmtyLine(nAt: Integer): Integer; {leere Trennzeile in FDates such}
  function GetLineRemain(nAt: Integer; s: string): string;  {liefert Zeilenrest}
 public
  constructor Create;
  destructor Destroy; override;
  property Server: String read FServer write FServer;              {POP3-Server}
  property Konto:  String read FKonto  write FKonto;          {Username/Account}
  property Pass:   String read FPass   write FPass;                   {Passwort}
  function CheckAccount: Integer;  {POP-Server kontaktieren und MailInfos holen}
  function CheckFromEml(sFNa: TFileName): Boolean;       {MailInfos lokal holen}
  property Checked: Boolean read FChecked; {Server wurde mit Erfolg kontaktiert}
  property Count:  Integer read FCount;      {Anzahl der Nachrichten in der Box}
  function GetSubject(n: Integer): String;              {Betreff der n-ten Mail}
  function GetFrom   (n: Integer): String;             {Absender der n-ten Mail}
  function GetTo     (n: Integer): String;            {Empfnger der n-ten Mail}
  function GetDate   (n: Integer): String;                {Datum der n-ten Mail}
  function GetSize   (n: Integer): String;                {Gre der n-ten Mail}
  function RetrMail(Nr: Integer): Integer;    {n-te Mail vom POP-Server abrufen}
  function DeleteMails(Nrs: String): Integer;     {Mails vom POP-Server lschen}
  function GetHeader: String;              {Header der zuletzt abgerufenen Mail}
  function GetMailText: String;        {plain-Text der zuletzt abgerufenen Mail}
  function GetAttachNames: String; {holt Anhang-Namen der zuletzt abgeruf. Mail}
  function StoreAsEml(sFNa: TFileName): Boolean; {speichert zuletzt abgerufene Mail im *.eml-Format}
  function LoadFromEml(sFNa: TFileName): Boolean;
  function StoreAttachment(Nr: Integer; sFNa: TFileName): Boolean; {speichert Anhang Nr der abgerufenen Mail}
end;

type TSMTPObject = class(TObject)  {Objekt fr SMTP}
 private
  FSMTPSock: TSocket;
  FServer:  String;                {Name des zu benutzenden SMTP-Servers}
  FMailTo:  String;                {Empfnger der e-Mail}
  FFrom:    String;                {e-Mail-Adresse des Absenders}
  FSender:  String;                {Absendername im Klartext}
  FSubject: String;                {Mail-Betreff}
  FSaveBySend, FSaveLn: Boolean;   {Speichern beim Senden}
  FSaveStream: TFileStream;        {Kopie des Sendedatenstromes}
  FSaveFileName: TFileName;        {Dateinamn des Sendedatenstromes}
  FContent: TStringList;           {Textinhalt der e-Mail in Zeilenform}
  FAttachments: TStringList;       {Dateinamensliste der e-MailAnhnge}
  function FContentLines: Integer;
  function SendMime(sFileNa: TFileName): Boolean;
  function SendLn(s: String): Integer;
  function Transmit: Boolean;
 public
  constructor Create;
  destructor Destroy; override;
  property Server: String read FServer write FServer;
  property MailTo: String read FMailTo write FMailTo;
  property From:   String read FFrom   write FFrom;
  property Sender: String read FSender write FSender;
  property Subject: String read FSubject write FSubject;
  property SaveBySend: Boolean read FSaveBySend write FSaveBySend;
  property SaveFile: TFileName read FSaveFileName write FSaveFileName;
  procedure SetContent(const TheText: String);
  procedure AddAttachment(const TheName: String);
  procedure ClearAttachments;
  function SendMail: Integer;
  function GetContentLine(n: Integer): String;
  property ContentLines: Integer read FContentLines;
end;

type PSMTPObject=^TSMTPObject; PPOP3Object=^TPOP3Object;

function HxCharToInt(c: Char): Byte;
function HtmlMaskLtGt(const s: String): String;
function DeBase64(const sIn: String): String;
procedure DeQuote(var sIn: String);
function DeQuode(const sIn: String): String;
function URLCode(const sIn: String): String;

implementation

uses Windows;

// allgemeine Hilfsroutinen ---------------------------------------------------

function HxCharToInt(c: Char): Byte;           {Dezimalwert eines Hex-Zeichens}
begin if c<='9' then HxCharToInt:=Ord(c)-48 else HxCharToInt:=Ord(UpCase(c))-55 end;

function HtmlMaskLtGt(const s: String): String; {"<" und ">" als &lt; und &gt;}
var t: String; i: Integer;
begin
 t:=s;
 i:=Pos('<',t);                                             {"<" herausfiltern}
 while i>0 do begin t[i]:='&'; Insert('lt;',t,i+1); i:=Pos('<',t) end;
 i:=Pos('>',t);                                             {">" herausfiltern}
 while i>0 do begin t[i]:='&'; Insert('gt;',t,i+1); i:=Pos('>',t) end;
 i:=Pos('"',t);                                             { "  herausfiltern}
 while i>0 do begin t[i]:='&'; Insert('quot;',t,i+1); i:=Pos('"',t) end;
 HtmlMaskLtGt:=t;
end;

function DeBase64(const sIn: String): String;       // BASE64-String decodieren
var j: Integer; c: Char; b: Byte; x, xc: Word; sT: String;
begin
 sT:=''; x:=0; xc:=0;
 for j:=1 to Length(sIn) do begin                           {ber alle Zeichen}
  c:=sIn[j];
  if (Ord(c)>20) and (c<>'=') then begin                {gltiges Mime-Zeichen}
   case c of
    '0'..'9': b:=Ord(c)+4;                                        { - '0' + 52}
    'A'..'Z': b:=Ord(c)-65;                                       { - 'A'}
    'a'..'z': b:=Ord(c)-71;                                       { - 'a' + 26}
    '+'     : b:=62;
    '/'     : b:=63;
   else b:=0 end;
   x:=x shl 6 + b; Inc(xc,6); if xc>=8 then begin Dec(xc,8); sT:=sT+Chr(x shr xc) end;
  end;
 end;
 DeBase64:=sT;
end;

procedure DeQuote(var sIn: String); {quoted-printabe Text in ANSI-Text konvert}
var p: Integer;
begin
 p:=Length(sIn)-2;
 repeat
  if sIn[p]='=' then begin
   if Copy(sIn,p+1,2)<>#13#10 then begin                  // kein Zeilenumbruch
    sIn[p]:=Chr((HxCharToInt(sIn[p+1]) shl 4)+HxCharToInt(sIn[p+2]));
    Delete(sIn,p+1,2);
   end else Delete(sIn,p,3);                               // bei Zeilenumbruch
  end;
  Dec(p);
 until p=0;
end;

function DeQuode(Const sIn: String): String; {=?-codierten Text in ANSI-Text konvert}
var t, s: String; nP1, nP2, nP3: Integer;
begin
 t:=sIn; nP1:=Pos('=?',t);
 while nP1<>0 do begin
  nP2:=Pos('?B?',t); nP3:=Pos('?Q?',t);
  if (nP3<>0) and ((nP2=0) or (nP3<nP2)) then begin                // Q-codiert
   Delete(t,nP1,nP3+3-nP1); nP3:=Pos('?=',t);
   if nP3<>0 then begin
    Delete(t,nP3,2); Dec(nP3,2);
    repeat
     if t[nP3]='=' then begin t[nP3]:=Chr((HxCharToInt(t[nP3+1]) shl 4)+HxCharToInt(t[nP3+2])); Delete(t,nP3+1,2) end;
     Dec(nP3);
    until nP3<nP1;
   end;
  end else if (nP2<>0) and ((nP3=0) or (nP2<nP3)) then begin       // B-codiert
   Delete(t,nP1,nP2+3-nP1); nP2:=Pos('?=',t);
   if nP2<>0 then begin
    Delete(t,nP2,2); s:=DeBase64(Copy(t,nP1,nP2-nP1)); Delete(t,nP1,nP2-nP1); Insert(s,t,nP1);
   end;
  end;
  nP1:=Pos('=?',t);
 end; // while
 DeQuode:=t;
end;

function URLCode(const sIn: String): String;          //String form-URL-kodieren
var s: String; i: Integer;
begin
 s:='';                                                     {noch berarbeiten!}
 for i:=1 to Length(sIn) do if ((sIn[i]>='@') and (sIn[i]<='z')) or ((sIn[i]>='0') and (sIn[i]<='9'))
  then s:=s+sIn[i] else s:=s+'%'+Format('%x',[Ord(sIn[i])]); URLCode:=s;
end;

{-----------------------------------------------------------------------------}
{ Die nchsten 3 Routinen stammen aus dem Projekt von Dennis Kalbhenn und
  Frank Ullmann, verffentlicht in der Zeitschrift PC-MAGAZIN Heft 2/98 S.200 }

function OpenConnection(Server:String; Port:Integer; var Erg:Integer): TSocket;
{erzeugt einen Socket und ffnet die Verbindung mit dem Server}
var HostEnt: PHostEnt; WSAData: TWSAData;
    MailSock: TSocket; SockAddr: TSockAddr; Addr: PInAddr;
begin
 OpenConnection:=0; Erg:=SOCK_NONET;
 if WSAStartup(MakeWord(1,1),WSAData)=0 then begin
  HostEnt:=GetHostByName(@Server[1]);
  if HostEnt<>nil then begin
   Addr:=Pointer(HostEnt.h_addr_list^);
   MailSock:=Socket(PF_INET, SOCK_STREAM, 0);
   if MailSock<>INVALID_SOCKET then begin
    SockAddr.sin_addr:=Addr^;
    SockAddr.sin_family:=PF_INET;
    SockAddr.sin_port:=HToNS(Port);
    if Connect(MailSock, SockAddr, SizeOf(SockAddr))=0 then begin
     Erg:=SOCK_CONNECT; OpenConnection:=MailSock;
    end else Erg:=SOCK_NOSERVER;
   end;
  end;
 end;
end;

procedure CloseConnection(ASock: TSocket);     {schliet die Socket-Verbindung}
begin CloseSocket(ASock); WSACleanup; end;

function SendString(ASock: TSocket; s: String): Integer; {in den Socket senden}
begin s:=s+#13#10; SendString:=Send(ASock, s[1], Length(s), 0) end;

//--- POP3-Routinen -----------------------------------------------------------

constructor TPOP3Object.Create;
begin
 inherited Create;
 FServer:=''; FKonto:=''; FPass:=''; FChecked:=false; FCount:=-1;
 FFroms:=TStringList.Create; FSubjects:=TStringList.Create;
 FDates:=TStringList.Create; FSizes:=TStringList.Create;
end;

destructor TPOP3Object.Destroy;
begin
 FSizes.Free; FDates.Free; FSubjects.Free; FFroms.Free;
 inherited Destroy
end;

function TPOP3Object.GetSubject(n: Integer): String;       {Betreff n ausgeben}
begin
 if FChecked and (n>=0) and (n<FCount) then Result:=DeQuode(FSubjects[n]) else Result:=''
end;

function TPOP3Object.GetFrom(n: Integer): String;         {Absender n ausgeben}
begin
 if FChecked and (n>=0) and (n<FCount) then Result:=DeQuode(FFroms[n]) else Result:=''
end;

function TPOP3Object.GetTo(n: Integer): String;            {Empfnger ausgeben}
begin
 if FChecked and (n>=0) and (n<FCount) then if FFroms.Count>n+1 then Result:=FFroms[n+1] else Result:=''
end;

function TPOP3Object.GetDate(n: Integer): String;   {Empfangs-Datum n ausgeben}
begin
 if FChecked and (n>=0) and (n<FCount) then Result:=FDates[n] else Result:=''
end;

function TPOP3Object.GetSize(n: Integer): String;       {Mailgrsse n ausgeben}
begin
 if FChecked and (n>=0) and (n<FCount) then Result:=FSizes[n] else Result:=''
end;

function TPOP3Object.GetEmtyLine(nAt: Integer): Integer;     {sucht Trennzeile}
var nMx, n: Integer;
begin
 n:=nAt; nMx:=FDates.Count;
 try
  while (n<nMx) and (FDates[n]<>'') do Inc(n);
  if FDates[n]='' then GetEmtyLine:=n else GetEmtyLine:=-1;
 except GetEmtyLine:=-1 end;
end;

function TPOP3Object.GetLineRemain(nAt: Integer; s: String): String;
{Schlsselbegriff s in e-Mail suchen und dahinterstehenden Zeilenrest ermitteln}
var nMx, n, i: Integer; t: String;
begin
 t:=UpperCase(s); n:=nAt; nMx:=FDates.Count;
 try
  while (n<nMx) and (Pos(t,UpperCase(FDates[n]))=0) do Inc(n);       {s suchen}
  i:=Pos(t,UpperCase(FDates[n]));
 except i:=0 end;
 if i>0 then GetLineRemain:=Trim(Copy(FDates[n],i+Length(t),255))
 else GetLineRemain:='';
end;

function TPOP3Object.FindLine(nAt: Integer; s: String): Integer;
{Zeile mit Anfangstext s in der empfangenen e-Mail suchen}
var nMx, n: Integer;
begin
 n:=nAt; nMx:=FDates.Count;
 try
  while (n<nMx) and (Pos(s,FDates[n])<>1) do Inc(n);
  if Pos(s,FDates[n])=1 then FindLine:=n else FindLine:=-1;
 except FindLine:=-1 end;
end;

function TPOP3Object.GetHeader: String; {liefert einige wichtige Header-Zeilen}
var t, s: String; i: Integer;
begin
 t:=''; i:=0;
 try
  while (i<FDates.Count) and (FDates[i]<>'') do begin {bis zur Header-Trennung}
   s:=UpperCase(Copy(FDates[i],1,3));
   if (s='FRO') or (s='TO:') or (s='DAT') or (s='SUB') or (s='RET') or
      (s='ENV')                           {hier beliebig ergnzen oder lschen}
   then t:=t+FDates[i]+#13#10;
   Inc(i)
  end;
  GetHeader:=DeQuode(t);
 except GetHeader:='fehlerhafter Nachrichtenkopf' end;
end;

function TPOP3Object.GetMailText: String;   {liefert den Textanteil der zuletzt
abgerufenen e-Mail (bis zur 3.Verschachtelungsebene bei multipart-Nachrichten)}
var nMx, nAt, nTx, i: Integer; t, sBound, sCt: String;
begin
 t:=''; nMx:=FDates.Count; nAt:=1;
 sBound:=GetLineRemain(0,'boundary=');          {Wert von boundary extrahieren}
 if sBound='' then                               {kein boundary, also nur Text}
  for i:=GetEmtyLine(1)+1 to nMx-1 do t:=t+FDates[i]+#13#10     {ab Trennzeile}
 else begin                                                   { multipart-Text}
  if sBound[1]='"' then sBound:=Copy(sBound,2,255);             {" abschneiden}
  if sBound[Length(sBound)]='"' then Delete(sBound,Length(sBound),1);
  sBound:='--'+sBound;                          {1. boundary-Muster extrahiert}
  nAt:=FindLine(1, sBound)+1;               {Content-Type nach boundary suchen}
  sCt:=Copy(GetLineRemain(nAt,'Content-Type:'),1,5);
  if sCt='text/' then begin                        {Content-Type ist: text/...}
   nTx:=GetEmtyLine(nAt)+1;                     {Trennzeile hinter 1. boundary}
   nMx:=FindLine(nTx, sBound)-1;                  {Zeile vor nchstem boundary}
   for i:=nTx to nMx do t:=t+FDates[i]+#13#10;
  end else if sCt='multi' then begin                       // nochmal multipart
   sBound:=GetLineRemain(nAt,'boundary=');   {Wert von 2. boundary extrahieren}
   if sBound<>'' then begin
    if sBound[1]='"' then sBound:=Copy(sBound,2,199);
    if sBound[Length(sBound)]='"' then Delete(sBound,Length(sBound),1);
    sBound:='--'+sBound;                        {2. boundary-Muster extrahiert}
    nAt:=FindLine(nAt,sBound)+1;         {Content-Type nach 2. boundary suchen}
    sCt:=Copy(GetLineRemain(nAt,'Content-Type:'),1,5);
    if sCt='text/' then begin
     nTx:=GetEmtyLine(nAt)+1;                   {Trennzeile hinter 2. boundary}
     nMx:=FindLine(nTx, sBound)-1;                {Zeile vor nchstem boundary}
     for i:=nTx to nMx do t:=t+FDates[i]+#13#10;
    end else if sCt='multi' then begin                      {nochmal multipart}
     sBound:=GetLineRemain(nAt,'boundary='); {Wert von 3. boundary extrahieren}
     if sBound<>'' then begin
      if sBound[1]='"' then sBound:=Copy(sBound,2,199);
      if sBound[Length(sBound)]='"' then Delete(sBound,Length(sBound),1);
      sBound:='--'+sBound;                      {3. boundary-Muster extrahiert}
      nAt:=FindLine(nAt,sBound)+1;       {Content-Type nach 3. boundary suchen}
      sCt:=Copy(GetLineRemain(nAt,'Content-Type:'),1,5);
      if sCt='text/' then begin
       nTx:=GetEmtyLine(nAt)+1;                 {Trennzeile hinter 3. boundary}
       nMx:=FindLine(nTx, sBound)-1;              {Zeile vor nchstem boundary}
       for i:=nTx to nMx do t:=t+FDates[i]+#13#10;
      end;
     end; // if 3. sBound <>''
    end; // if sCt='multi'
   end; // if 2. sBound <>''
  end; // if sCt='multi'
 end; // if 1. sBound <>''
 if LowerCase(GetLineRemain(nAt,'-Encoding:'))='quoted-printable' then DeQuote(t);
 GetMailText:=t;
end;

function TPOP3Object.GetAttachNames: String;  {Namen der Attachments ermitteln}
var nAt, nEnd: Integer; s, t, sBound: String;
begin
 s:=''; t:=GetLineRemain(0,'Content-Type:');     {1-ten Content-Type bestimmen}
 sBound:=GetLineRemain(0,'boundary=');          {Wert von boundary extrahieren}
 if (sBound<>'') and (LowerCase(Copy(t,1,15))='multipart/mixed') then begin
  if sBound[1]='"' then sBound:=Copy(sBound,2,199);
  if sBound[Length(sBound)]='"' then Delete(sBound,Length(sBound),1);
  sBound:='--'+sBound;                             {boundary fertig extrahiert}
  nAt:=FindLine(FindLine(1, sBound)+1, sBound);         {2-tes boundary suchen}
  nEnd:=FDates.Count-20; if nEnd<nAt then nEnd:=nAt;
  nEnd:=FindLine(nEnd, sBound+'--');           {abschlieendes boundary suchen}
  while nAt<nEnd do begin               {vom 2-ten bis abschlieenden boundary}
   t:=GetLineRemain(nAt,'name='); if t='' then t:='unbenannt';
   if t[1]='"' then t:=Copy(t,2,199); if t[Length(t)]='"' then Delete(t,Length(t),1);
   s:=s+DeQuode(t)+#13#10;                                {Dateinamen anhngen}
   nAt:=FindLine(nAt+1, sBound);                     {nchstes Boundary suchen}
  end;
 end;
 GetAttachNames:=s;
end;

function TPOP3Object.StoreAsEml(sFNa: TFileName): Boolean; {speichert zuletzt abgerufene e-Mail}
begin
 try FDates.SaveToFile(sFNa); StoreAsEml:=true except StoreAsEml:=false end;
end;

function TPOP3Object.LoadFromEml(sFNa: TFileName): Boolean;
begin
 try FDates.LoadFromFile(sFNa); LoadFromEml:=true except LoadFromEml:=false end;
end;

function TPOP3Object.StoreAttachment(Nr: Integer; sFNa: TFileName): Boolean;
{speichen des Anhangs Nr der zuletzt abgerufenen e-Mail unter dem Namen sFNa}
var OutStr: TFileStream; sBound, sEncod, sLn: String;
    i, j, nAt, nEnd: Integer; c: Char; b: Byte; x, xc, y: Word;
begin
 sBound:=GetLineRemain(0,'boundary=');          {Wert von boundary extrahieren}
 sEncod:=GetLineRemain(0,'Content-Type:');
 if (sBound<>'') and (LowerCase(Copy(sEncod,1,15))='multipart/mixed') then begin
  if sBound[1]='"' then sBound:=Copy(sBound,2,199);
  if sBound[Length(sBound)]='"' then Delete(sBound,Length(sBound),1);
  sBound:='--'+sBound;
  nAt:=FindLine(1, sBound);              {1-tes boundary vor plain-Text suchen}
  for i:=0 to Nr do nAt:=FindLine(nAt+1, sBound);      {Nr-tes boundary suchen}
  sEncod:=GetLineRemain(nAt,'-Transfer-Encoding:');
  nAt:=GetEmtyLine(nAt)+1;                     {Trennzeile im Nr-ten Abschnitt}
  nEnd:=FindLine(nAt+1, sBound)-1;                          {nchstes boundary}
  try
   OutStr:=TFileStream.Create(sFNa, fmCreate);
   try
    if UpperCase(sEncod)='BASE64' then begin         {mime-dekodiert speichern}
     x:=0; xc:=0;
     for i:=nAt to nEnd do begin              {ber alle Zeilen des Abschnitts}
      sLn:=FDates[i];
      for j:=1 to Length(sLn) do begin            {ber alle Zeichen der Zeile}
       c:=sLn[j];
       if (Ord(c)>20) and (c<>'=') then begin           {gltiges Mime-Zeichen}
        case c of
         '0'..'9': b:=Ord(c)+4;                                   { - '0' + 52}
         'A'..'Z': b:=Ord(c)-65;                                  { - 'A'}
         'a'..'z': b:=Ord(c)-71;                                  { - 'a' + 26}
         '+'     : b:=62;
         '/'     : b:=63;
        else b:=0 end;
        x:=x shl 6 + b; Inc(xc,6);
        if xc>=8 then begin Dec(xc,8); y:=x shr xc; OutStr.Write(y,1); end;
       end;
      end; // alle j Zeichen
     end; //alle i Zeilen
    end else begin     { unbearbeitet speichern (evt. noch UUencoded prfen??)}
     for i:=nAt to nEnd do begin
      sLn:=FDates[i]+#13#10; OutStr.Write(sLn[1],Length(sLn));
     end;
    end;
    StoreAttachment:=true;                                    {ist gespeichert}
   finally OutStr.Free end;
  except StoreAttachment:=false end;
 end else StoreAttachment:=false;
end;

function TPOP3Object.CheckFromEml(sFNa: TFileName): Boolean;
{ermittelt Kopfangaben aus einer im "*.eml"-Format gespeicherten Nachricht}
var Dat: Text; sLn, s, t, sF, sD, sS, sT: String; i: Integer;
begin
 t:=''; AssignFile(Dat, sFNa);
 try
  Reset(Dat); repeat ReadLn(Dat,sLn); FDates.Add(sLn) until sLn=''; CloseFile(Dat);
  for i:=0 to FDates.Count-1 do begin
   s:=FDates[i]; t:=UpperCase(Copy(s,1,6));
   if t='FROM: ' then sF:=Copy(s,7,199);
   if t='DATE: ' then sD:=Copy(s,7,199);
   if t='SUBJEC' then sS:=Copy(s,10,199);
   if Copy(t,1,4)='TO: ' then sT:=Copy(s,5,199);
  end;
  FDates.Clear; FFroms.Clear; FSubjects.Clear; FSizes.Clear; FSizes.Add('0');
  FFroms.Add(sF); FFroms.Add(sT); FSubjects.Add(sS); FDates.Add(sD); FCount:=1; FChecked:=true;
  CheckFromEml:=true;
 except CheckFromEml:=false end;
end;

function TPOP3Object.CheckAccount: Integer;
{kontaktiert den POP-Server und holt wesentliche Kopfdaten zu allen e-Mails ab}
var Header: TStringList;
    nErg, i, j: Integer; r: array[0..4095] of Byte; s, t, sF, sS, sD: String;
begin
 FChecked:=false;
 if (Length(FServer)>3)and(Length(FKonto)>0)and(Length(FPass)>0) then begin
  try
   nErg:=SOCK_NONET;
   FPopSock:=OpenConnection(FServer, IPPORT_POP3, nErg);
   if FPopSock>0 then begin
    Recv(FPopSock,r[0],1024,0);      {Besttigung zum Connect, nicht auswerten}
    SendString(FPopSock,'USER '+FKonto); Recv(FPopSock,r[0],1024,0);
    if r[0]=Ord('+') then begin                                       {USER OK}
     SendString(FPopSock,'PASS '+FPass); Recv(FPopSock,r[0],1024,0);
     if r[0]=Ord('+') then begin                                      {PASS OK}
      nErg:=POP3_MAILOK; FChecked:=true;
      SendString(FPopSock,'STAT'); Recv(FPopSock,r[0],1024,0);
      if r[0]=Ord('+') then begin                                     {STAT OK}
       s:=''; i:=4; while r[i]<>$20 do begin s:=s+Chr(r[i]); Inc(i) end;
       try FCount:=StrToInt(s) except FCount:=0 end;
       if FCount>0 then begin                                  {Nachrichten da}
        SendString(FPopSock,'LIST'); t:='';
        repeat
         i:=Recv(FPopSock,r[0],1024,0); r[i]:=0; t:=t+PChar(@r[0]);
        until Pos(#10'.',t)>0;
        i:=1;
        for j:=1 to FCount do begin                     {ber alle Nachrichten}
         while t[i]<>#10 do Inc(i);                              {LF/CR suchen}
         while t[i]<>' ' do Inc(i);              {1. SPACE in der Zeile suchen}
         s:=''; while t[i]>=' ' do begin s:=s+t[i]; Inc(i) end;
         FSizes.Add(s);                                       {Lnge eintragen}
        end;
        Header:=TStringList.Create;
        for j:=1 to FCount do begin                     {ber alle Nachrichten}
         SendString(FPopSock,Format('TOP %d 0',[j]));
         t:=''; sF:='?'; sS:='?'; sD:='?';
         repeat
          i:=Recv(FPopSock,r[0],4095,0); r[i]:=0; t:=t+PChar(@r[0]);
         until Pos(#10'.'#13,t)>0;
         Header.Text:=t;
         for i:=0 to Header.Count-1 do begin
          s:=Header[i]; t:=UpperCase(Copy(s,1,6));
          if t='FROM: ' then sF:=Copy(s,7,199);
          if t='DATE: ' then sD:=Copy(s,7,199);
          if t='SUBJEC' then sS:=Copy(s,10,199);
         end;
         FFroms.Add(sF); FSubjects.Add(sS); FDates.Add(sD);
        end;
        Header.Free;
       end; // FCount>0
      end; // STAT +OK
     end else nErg:=POP3_NOPASS;
    end else nErg:=POP3_NOKONTO;
    SendString(FPopSock,'QUIT');
    CloseConnection(FPopSock);
   end else nErg:=SOCK_NOSERVER;
  except nErg:=SOCK_NONET end;
  CheckAccount:=nErg;
 end else CheckAccount:=POP3_NOPARAMS;
end;

function TPOP3Object.RetrMail(Nr: Integer): Integer; {ruft Mail Nr komplett ab}
var nErg, i: Integer; r: array[0..4095] of Byte; sBuf: String;
begin
 if (Length(FServer)>3) and (Length(FKonto)>0) and (Length(FPass)>0) then begin
  try
   nErg:=SOCK_NONET;
   FPopSock:=OpenConnection(FServer, IPPORT_POP3, nErg);
   if FPopSock>0 then begin
    Recv(FPopSock,r[0],1024,0);      {Besttigung zum Connect, nicht auswerten}
    SendString(FPopSock,'USER '+FKonto); Recv(FPopSock,r[0],1024,0);
    if r[0]=Ord('+') then begin                                       {USER OK}
     SendString(FPopSock,'PASS '+FPass); Recv(FPopSock,r[0],1024,0);
     if r[0]=Ord('+') then begin                                      {PASS OK}
      nErg:=POP3_MAILOK;
      SendString(FPopSock,'STAT'); Recv(FPopSock,r[0],1024,0);
      if r[0]=Ord('+') then begin                                     {STAT OK}
       SendString(FPopSock,Format('RETR %d',[Nr])); sBuf:='';
       repeat
        i:=Recv(FPopSock,r[0],4095,0); r[i]:=0; sBuf:=sBuf+PChar(@r[0])
       until Pos(#10'.'#13,Copy(sBuf,Length(sBuf)-8,8))>0;
       FDates.Text:=sBuf; sBuf:=''; FChecked:=false;
       if FDates.Count>0 then begin
        i:=FDates.Count-5; if i<0 then i:=0;
        while i<FDates.Count do
        begin if FDates[i]='.' then FDates.Delete(i) else Inc(i) end;
        FDates.Delete(0);                      {Zeile "+OK xxx octets" lschen}
       end;
      end;
     end else nErg:=POP3_NOPASS;
    end else nErg:=POP3_NOKONTO;
    SendString(FPopSock,'QUIT');
    CloseConnection(FPopSock);
   end else nErg:=SOCK_NOSERVER;;
  except nErg:=SOCK_NONET end;
  RetrMail:=nErg;
 end else RetrMail:=POP3_NOPARAMS;
end;

function TPOP3Object.DeleteMails(Nrs: String): Integer;
{Mails mit Nummern laut Liste Nrs vom POP-Server lschen}
var nErg, i: Integer; r: array[0..1023] of Byte; Mails: TStringList;
begin
 if (Length(FServer)>3) and (Length(FKonto)>0) and (Length(FPass)>0) then begin
  try
   nErg:=SOCK_NONET;
   FPopSock:=OpenConnection(FServer, IPPORT_POP3, nErg);
   if FPopSock>0 then begin
    Recv(FPopSock,r[0],1024,0);      {Besttigung zum Connect, nicht auswerten}
    SendString(FPopSock,'USER '+FKonto); Recv(FPopSock,r[0],1024,0);
    if r[0]=Ord('+') then begin                                       {USER OK}
     SendString(FPopSock,'PASS '+FPass); Recv(FPopSock,r[0],1024,0);
     if r[0]=Ord('+') then begin                                      {PASS OK}
      nErg:=POP3_MAILOK;
      SendString(FPopSock,'STAT'); Recv(FPopSock,r[0],1024,0);
      if r[0]=Ord('+') then begin                                     {STAT OK}
       Mails:=TStringList.Create;
       Mails.Text:=Nrs;
       for i:=0 to Mails.Count-1 do begin
        SendString(FPopSock,'DELE '+Mails[i]); Recv(FPopSock,r[0],1024,0);
       end;
       Mails.Free;
      end;
     end else nErg:=POP3_NOPASS;
    end else nErg:=POP3_NOKONTO;
    SendString(FPopSock,'QUIT');
    CloseConnection(FPopSock);
   end else nErg:=SOCK_NOSERVER;;
  except nErg:=SOCK_NONET end;
  DeleteMails:=nErg;
 end else DeleteMails:=POP3_NOPARAMS;
end;

//--- SMTP-Routinen -----------------------------------------------------------

constructor TSMTPObject.Create;
begin
 inherited Create;
 FServer:=''; FMailTo:=''; FFrom:=''; FSender:=''; FSubject:=''; FSaveBySend:=false; FSaveLn:=false;
 FContent:=TStringList.Create; FAttachments:=TStringList.Create
end;

destructor TSMTPObject.Destroy;
begin FAttachments.Free; FContent.Free; inherited Destroy end;

function TSMTPObject.SendLn(s: String): Integer;
begin
 s:=s+#13#10; SendLn:=Send(FSMTPSock, s[1], Length(s), 0);
 if FSaveLn then FSaveStream.Write(s[1], Length(s));
end;

function TSMTPObject.FContentLines: Integer;
begin FContentLines:=FContent.Count end;

function TSMTPObject.GetContentLine(n: Integer): String;
begin
 if (n>=0) and (n<FContent.Count) then Result:=FContent[n] else Result:='';
end;

procedure TSMTPObject.SetContent(const TheText: String);
begin FContent.Text:=TheText end;

procedure TSMTPObject.AddAttachment(const TheName: String);
begin FAttachments.Add(TheName) end;

procedure TSMTPObject.ClearAttachments;
begin FAttachments.Clear end;

function TSMTPObject.SendMail: Integer;                         {Hauptprozedur}
var nErg: Integer;
begin
 if (Length(FServer)>3)and(Length(FFrom)>5)and(Length(FMailTo)>5) then begin
  if FSubject='' then FSubject:='no subject';
  try
   FSMTPSock:=OpenConnection(FServer, IPPORT_SMTP, nErg);
   if FSMTPSock>0 then begin
    if Transmit then nErg:=SMTP_MAILOK else nErg:=SMTP_TRANSERR;
    CloseConnection(FSMTPSock);
   end else nErg:=SOCK_NOSERVER;
  except nErg:=SOCK_NONET end;
  SendMail:=nErg;
 end else SendMail:=SMTP_NOPARAMS;
end;

function TSMTPObject.Transmit: Boolean;
{bertrgt eine e-Mail an den Mail-Server gem Protokoll nach RFC 821}
var i: Integer; r: array[0..1023] of Byte; sTo, sBd, sFNa: String; bMime: Boolean;
begin
 Transmit:=false;
 Recv(FSMTPSock,r[0],1024,0);  {Bestt. zum Connect lesen aber nicht auswerten}
 SendLn('HELO '+FServer);
 Recv(FSMTPSock,r[0],1024,0); {Besttigung zum HELO lesen aber nicht auswerten}
 SendLn('MAIL FROM:<'+FFrom+'>');
 Recv(FSMTPSock,r[0],1024,0);
 sTo:=FMailTo;
 if Pos('<',sTo)>0 then begin                {Empfnger-e-Mail herausschneiden}
  sTo:=Copy(sTo,Pos('<',sTo)+1,99);
  if Pos('>',sTo)>0 then sTo:=Copy(sTo,1,Pos('>',sTo)-1);
 end;
 SendLn('RCPT TO:<'+sTo+'>');
 Recv(FSMTPSock,r[0],1024,0);
 SendLn('DATA');
 Recv(FSMTPSock,r[0],1024,0);
 if FSaveBySend and (FSaveFileName<>'') then try
  FSaveStream:=TFileStream.Create(FSaveFileName,fmCreate or fmOpenWrite); FSaveLn:=true;
 except FSaveLn:=false end;
 SendLn('From: '+FSender+' <'+FFrom+'>');
 SendLn('To: '+FMailTo);
 SendLn('Subject: '+FSubject);
 SendLn(FormatDateTime('"Date:" ddd, d mmm yyyy hh:nn:ss',Now));
 SendLn('X-Mailer: WebMail-Interface (c) by J. Hummel (www.hummel1.de)');
 bMime:=FAttachments.Count<>0;
 if bMime then begin
  SendLn('MIME-Version: 1.0');
  sBd:=FormatDateTime('"-----_NEXT_"hhnnss_ddmmyy"_WM"',Now);
  SendLn('Content-Type: multipart/mixed; '#13#10' boundary="'+sBd+'"'); sBd:='--'+sBd;
  SendLn(#13#10'mehrteilige Nachricht im MIME-Format'#13#10#13#10+sBd);
 end;
 SendLn('Content-Type: text/plain; charset=iso-8859-1');
 SendLn('Content-Transfer-Encoding: 8bit'#13#10);         {Leerzeile mu sein!}
 for i:=0 to FContent.Count-1 do SendLn(FContent[i]);                    {Text}
 if bMime then begin
  for i:=0 to FAttachments.Count-1 do begin
   sFNa:=FAttachments[i];
   SendLn(#13#10+sBd);
   SendLn('Content-Type: application/octet-stream; name="'+ExtractFileName(sFNa)+'"');
   SendLn('Content-Transfer-Encoding: base64');
   SendLn('Content-Disposition: attachment; filename="'+ExtractFileName(sFNa)+'"'#13#10);
   SendMime(sFNa);
   SendLn('');
  end;
  SendLn(#13#10+sBd+'--');
 end;
 if FSaveLn then begin FSaveLn:=false; FSaveStream.Free end;
 SendLn(#13#10'.');                             {Ende-Token der Daten anhngen}
 Recv(FSMTPSock,r[0],1024,0);         {Besttigung lautet bei Erfolg: "250 OK"}
 if (r[0]=Ord('2'))and(r[1]=Ord('5'))and(r[2]=Ord('0')) then Transmit:=true;
 SendLn('QUIT');
 Recv(FSMTPSock,r[0],1024,0);
end;

function TSMTPObject.SendMime(sFileNa: TFileName): Boolean;
const mc='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
var Source: TFileStream; b, br: Byte; w: Word; l: Longint; s: String;
begin
 s:='';
 try
  Source:=TFileStream.Create(sFileNa,fmOpenRead);
  repeat
   if Length(s)>64 then begin SendLn(s); s:='' end;
   b:=0; w:=0;
   br:=Source.Read(b,1); br:=br+Source.Read(w,2); l:=makeLong(swap(w),word(b));
   s:=s+mc[((l shr 18) and 63)+1]+mc[((l shr 12) and 63)+1]+mc[((l shr 6) and 63)+1]+mc[(l and 63)+1];
  until br<3;
  if br=1 then s:=copy(s,1,length(s)-2)+'==';  if br=2 then s:=copy(s,1,length(s)-1)+'=';
  if s<>'' then SendLn(s);
  Source.Free;  SendMime:=true;
 except SendMime:=false end;
end;

end.
